#%%
import io
import base64

from dash import Dash, dcc, html, Input, Output, no_update, callback
import plotly.express as px

import dash_bootstrap_components as dbc
import pandas as pd
import sys
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import seaborn as sns
import os
import re
import glob
import random
from sklearn import preprocessing
from tqdm import tqdm
import plotly.express as px
from PIL import Image
import plotly.figure_factory as ff

dataPath = r'D:\Gil\images\pex_project\02142024\single_images\imageSequence\data'
IMAGE_FOLDER = r'D:\Gil\images\pex_project\02142024\single_images\imageSequence\images'
df = pd.read_csv(os.path.join(dataPath,'imageseq_data.csv'))
# float_cols = df.select_dtypes(include=['float']).columns
# df[float_cols] = df[float_cols].apply(lambda x: np.round(x, 3))
# np.random.seed(42423)
# df = df.sample(frac=1).reset_index(drop=True)
# #df['name'] = df['name'].apply(lambda x: f"{x}.png")
# first_image_name = df.iloc[0]['name']
# first_image_path = f"{IMAGE_FOLDER}/{first_image_name}"
# savedContainer = {'name':[], 'ratio':[], 'intensity':[], 'sd':[], 'maskArea':[], 'label':[]}
# image_names = [re.sub("_.*","",name) for name in set(df.name.tolist())]
#%%
hist_data = [df.loc[df['label']==1,'ratio'].values, df.loc[df['label']==0,'ratio'].values]
image_name_txt = [df.loc[df['label']==1,'name'].values, df.loc[df['label']==0,'name'].values]
group_labels = ['phno', 'wt']
fig = ff.create_distplot(hist_data, group_labels, bin_size=len(hist_data)/10,rug_text = image_name_txt)
fig.show()


#intensity
#%% 
hist_data = [df.loc[df['label']==1,'intensity'].values, df.loc[df['label']==0,'intensity'].values]
image_name_txt = [df.loc[df['label']==1,'name'].values, df.loc[df['label']==0,'name'].values]
group_labels = ['phno', 'wt']

fig = ff.create_distplot(hist_data, group_labels, bin_size=len(hist_data)/2,rug_text = image_name_txt)
fig.show()

#sd
#%% 
hist_data = [df.loc[df['label']==1,'sd'].values, df.loc[df['label']==0,'sd'].values]
image_name_txt = [df.loc[df['label']==1,'name'].values, df.loc[df['label']==0,'name'].values]
group_labels = ['phno', 'wt']

fig = ff.create_distplot(hist_data, group_labels, bin_size=len(hist_data)/2,rug_text = image_name_txt)
fig.show()





import matplotlib.pyplot as plt
import arviz as az
import pymc as pm
import os
from matplotlib.backends.backend_pdf import PdfPages
from sklearn import metrics
from sklearn.metrics import precision_recall_fscore_support
import numpy as np
RANDOM_SEED = 8927
np.random.seed(RANDOM_SEED)
import pandas
import string
import cv2
import pdb
import random
import skimage
import seaborn as sns
import os
import pandas as pd
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
from IPython.display import clear_output
from matplotlib.backends.backend_pdf import PdfPages
from AIPyS.segmentation.cellpose.StackObjects_cellpose import StackObjects_cellpose

class Baysian_training(StackObjects_cellpose):
    '''
    The function `TrainBayesImage` requires a labeled table as input, which includes the columns: `name`, `ratio`, `maskArea`, `label`. 
    It outputs a Bayesian model. The input data is created using the `GranularityDataGen.py` function and manually labeled through a labeling dashboard application. The final output of the function is a table detailing the performance of the model.
    Parameters
    ------------
    dataFileName : string
        csv file name e.g. 'imageseq_data.csv'
    dataPath : string
        path to csv file
    imW: int
        E.G. 10
    IMh: int
        E.G. 5
    thold: float 
        0 to 1
    '''
    def __init__(self, fractionData,areaSel, dataPath, imW, imH, thold, *args , **kwargs):
        self.fractionData = fractionData
        self.areaSel = areaSel
        self.dataPath = dataPath
        self.imW = imW
        self.imH = imH
        self.thold = thold
        super().__init__(*args, **kwargs)
        self.df = self.loadCSV()
        #self.intercept, self.slope = self.bayesModelTraining()
        

    def loadCSV(self):
        return pd.read_csv(os.path.join(self.dataPath,'imageseq_data.csv'))
    
    def matchSize(self):
        '''
        Parameters
        __________
        fractionData: int
            work on fraction of the data for training
        areaSel: int
            cutoff for area
        '''
        df_label = self.df
        if self.areaSel > 0:
            df_label = df_label.loc[df_label['maskArea'] > self.areaSel,:].reset_index(drop=True)
        if self.fractionData > 0:
            if self.fractionData > len(df_label):
                df_label = df_label.sample(n=len(df_label)-1, replace=False).reset_index(drop=True)
            else:
                df_label = df_label.sample(n=self.fractionData, replace=False).reset_index(drop=True)
        if self.areaSel < 0 and self.fractionData < 0:
            OneNum = len(df_label.loc[df_label["label"]==1])
            ind_pheno_list = df_label.loc[df_label["label"]==1,:].index.to_list()
            ind_WT_list = df_label.loc[df_label["label"]==0,:].index.to_list()
            random.shuffle(ind_WT_list)
            ind_WT_list_sel= [ind_WT_list[i] for i in range(OneNum)]
            df_label = df_label.iloc[ind_WT_list_sel+ind_pheno_list,:]
        return df_label
        
    
    def pairDistribution(self):
        df_label = self.matchSize()
        colSel = ['name','ratio','maskArea','label']
        df_label_pair = df_label.loc[:,colSel]
        label_map = {0: "WT", 1: "Pheno"}
        df_label_pair['label'] = df_label_pair['label'].map(label_map)
        palette1 = sns.color_palette("colorblind",4)
        custom_palette = {"WT": palette1[2], "Pheno": palette1[3]}
        plt.rcParams['figure.dpi'] = 150
        plt.rcParams['font.family'] = ['serif']
        plt.rcParams['font.size'] = 12
        plt.rcParams['axes.labelsize'] = 12
        plt.rcParams['axes.titlesize'] = 12
        plt.rcParams['xtick.labelsize'] = 12
        plt.rcParams['ytick.labelsize'] = 12
        
        g = sns.pairplot(df_label_pair, hue="label", kind="kde", palette=custom_palette)
        g.fig.set_size_inches(self.imW, self.imH)  # Set the figure size (width, height) in inches
        with PdfPages(os.path.join(self.outPath,'pairplot_output.pdf')) as pdf:
            pdf.savefig(g.fig)  # Save the figure `g.fig` into the PDF file
            plt.close(g.fig) 
    
    def classify(self,Int,trace):
        '''
        Parameters
        ------------ 
        Int array of intensities
        thold float
        trace pymc object
        '''
        mu = trace.posterior['a'].mean(dim=("chain", "draw")).values + trace.posterior['b'].mean(dim=("chain", "draw")).values * Int
        prob = 1 / (1 + np.exp(-mu))
        return prob, prob > self.thold
    
    def bayesModelTraining(self):
        df_label = self.matchSize()
        rate = df_label['ratio'].values
        y_0 = df_label['label'].values
        with pm.Model() as model_logistic_basic:
            a = pm.Normal('a', 0, 2)
            b = pm.Normal('b', 0, 2)
            mu = a + pm.math.dot(rate, b)
            theta = pm.Deterministic('theta', 1 / (1 + pm.math.exp(-mu)))
            bd = pm.Deterministic('bd', -a / b)
            yl = pm.Bernoulli('yl', theta, observed=y_0)
            trace = pm.sample(1000, tune=2000, target_accept=0.99, random_seed=RANDOM_SEED)
        prob, prediction = self.classify(Int = rate, trace = trace)
        y_true = y_0
        y_pred = np.where(prediction == True, 1, 0)
        performance = precision_recall_fscore_support(y_true, y_pred, average='macro')
        with PdfPages(os.path.join(self.outPath,'BayesReport.pdf')) as pdf:
            plt.figure(figsize=(3, 3))
            plt.title('Trace Plot')
            az.plot_trace(trace, figsize=(12, 6), compact=True)
            az.plot_trace(trace, figsize=(12, 6), compact=True)
            pdf.savefig()
            plt.close()

            idx = np.argsort(rate)
            theta = trace.posterior['theta'].mean(dim=("chain", "draw")).values
            plt.figure(figsize=(3, 3))
            plt.title('Boundary plot')
            plt.plot(rate[idx], theta[idx], color='b', lw=3)
            plt.axvline(trace.posterior['bd'].mean(), ymax=1, color='r')
            bd_hdi = pm.hdi(trace.posterior['bd'])
            bd_low = pm.hdi(trace.posterior['bd']).sel(hdi='lower')['bd'].values
            bd_high = pm.hdi(trace.posterior['bd']).sel(hdi='higher')['bd'].values
            plt.fill_betweenx([0, 1], bd_low, bd_high, color='r')
            plt.plot(rate, y_0, 'o', color='k')
            pdf.savefig()
            plt.close()

            plt.figure(figsize=(3, 3))
            plt.title('Performance')
            confusion_matrix = metrics.confusion_matrix(np.array(df_label['label'].values, dtype=int),np.where(prediction, 1, 0))
            cm_display = metrics.ConfusionMatrixDisplay(confusion_matrix=confusion_matrix, display_labels=[False, True])
            cm_display.plot()
            plt.text(0, 0.5, "Precision :{}".format(np.round(performance[0],2)), fontsize=10, transform=plt.gcf().transFigure)
            plt.text(0, 0.4, "Recall :{}".format(np.round(performance[1],2)), fontsize=10, transform=plt.gcf().transFigure)
            plt.text(0, 0.3, "F1 score :{}".format(np.round(performance[2],2)), fontsize=10, transform=plt.gcf().transFigure)
            plt.text(0, 0.2, "a :{}".format(np.round(trace.posterior['a'].mean().values, 2)), fontsize=10,transform=plt.gcf().transFigure)
            plt.text(0, 0.1, "b :{}".format(np.round(trace.posterior['b'].mean().values, 2)), fontsize=10,transform=plt.gcf().transFigure)
            pdf.savefig()
            plt.close()
        intercept = trace.posterior['a'].mean().values
        slope = trace.posterior['b'].mean().values 
        return  intercept, slope 
            
            
        
        
        




    
    


import pandas as pd
import sys
import numpy as np
np.set_printoptions(threshold=sys.maxsize)
from tqdm import tqdm

import matplotlib.pyplot as plt
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
import seaborn as sns
#sns.set()
import arviz as az
import pymc3 as pm
print(pm.__version__)
import theano.tensor as tt
import patsy

import os
import re
import glob
import random
# import plotnine
from sklearn import preprocessing
from tqdm import tqdm

import plotly.express as px
from matplotlib.backends.backend_pdf import PdfPages
from sklearn import metrics
from sklearn.metrics import precision_recall_fscore_support

from skimage import measure, restoration,morphology
from skimage import io, filters, measure, color, img_as_ubyte
from skimage.draw import disk
from skimage import measure, restoration,morphology

RANDOM_SEED = 8927
np.random.seed(RANDOM_SEED)
from scipy.special import expit as logistic

class BayesianModels():
    def __init__(self, df):
        self.df = df

    def getValues(self,geneSelect):
        valueInput = self.df.loc[self.df.Gene.str.contains(geneSelect),'realFoldChange'].values
        return valueInput

    def PossionRegression(self):
        RANDOM_SEED = 8927
        hitMap = {"Gene":[],"Target_stdLog2Raw":[],"non_stdLog2Raw":[],"Target_Posterior":[] ,"non_Posterior":[] }
        # standertized log2 fold change
        nonStdLog =  self.getValues(geneSelect = 'non')
        Genes = self.df.Gene.unique().tolist()
        Genes.remove('non')
        for gene in tqdm(Genes):
            # values of selected gene
            CurrValueStdLog = self.getValues(geneSelect =gene)
            GeneID = np.hstack([np.repeat(0, len(nonStdLog)), np.repeat(1, len(CurrValueStdLog))])
            y_all = np.hstack([nonStdLog,CurrValueStdLog])
            with pm.Model() as m11_12:
                a = pm.Normal("a", 0.0, 2.0)
                b = pm.Normal("b", 0.0, 2.0)
                lam = pm.math.exp(a + b * GeneID)
                y = pm.Poisson("y", lam, observed=y_all)
                trace = pm.sample(1000, tune=2000, random_seed=RANDOM_SEED)
            x = 1
            non = np.exp(trace["posterior"]["a"])
            target = np.exp(trace["posterior"]["a"] + trace["posterior"]["b"])
            hitMap['Gene'].append(gene)
            hitMap['Target_stdLog2Raw'].append(np.mean(CurrValueStdLog))
            hitMap['non_stdLog2Raw'].append(np.mean(nonStdLog))
            hitMap['Target_Posterior'].append(target)
            hitMap['non_Posterior'].append(non)



import numpy as np
from skimage import measure, morphology
from skimage.transform import resize
import pandas as pd
from random import randint

class GRANULARITY():
    def __init__(self,mask,image):
        self.mask = mask
        self.image = image

    def outlierRemoval(self):
        pass

    def measure_properties(self, input_image, input_mask, prop_names):
        '''
        :param: input_image (img), 8 bit greyscale image
        :param: mask (img) 32bit mask
        :param: prop_names (list), features array
        :return: skit image property table propriety table
        '''

        def sd_intensity(regionmask, intensity_image):
            return np.std(intensity_image[regionmask])

        def pixelcount(regionmask):
            return np.sum(regionmask)

        def mean_int(regionmask, intensity_image):
            return np.mean(intensity_image[regionmask])

        if prop_names is None:
            raise ValueError("features array is missing or not of type list")

        table_prop = measure.regionprops_table(
            label_image=input_mask, intensity_image=input_image, properties=prop_names,
            extra_properties=(sd_intensity, pixelcount, mean_int))
        return table_prop

    def featuresTable(self,features =["label"]):
        '''
        :param: features (list), could be a list or tuple
        :return: skit image property table and pandas DF of image intensity at the smallest kernel
        '''
        prop_names = (
            "label",
            "centroid",
            "area",
            "eccentricity",
            "euler_number",
            "extent",
            "feret_diameter_max",
            "inertia_tensor",
            "inertia_tensor_eigvals",
            "moments",
            "moments_central",
            "moments_hu",
            "moments_normalized",
            "orientation",
            "perimeter",
            "perimeter_crofton",
            "slice",
            "solidity")
        # incase fetures are type of list then no need to do the loop
        if not(features[0] in prop_names):
            mesg = "select from the following features:{}".format(prop_names)
            raise ValueError(mesg)
        featuresSel = [feature for feature in prop_names if feature in features]
        tableOut = pd.DataFrame(self.measure_properties(input_image = self.image,input_mask = self.mask ,prop_names = featuresSel))
        # initiate table signal time zero, first column per label
        tableInit = pd.DataFrame(tableOut['mean_int'].tolist())
        return tableOut, tableInit

    def openingOperation(self,kernel):
        '''
        :param: kernel (int), size of filter kernel
        :return: opening operation image
        '''
        selem = morphology.disk(kernel, dtype=bool)
        eros_pix = morphology.erosion(self.image, footprint=selem)
        imageOpen = morphology.dilation(eros_pix, footprint=selem)
        return imageOpen

    def loopLabelimage(self,start_kernel = 2 , end_karnel = 80, kernel_size=20 , fetureLabel = ["label"], deploy = False ):
        '''
        :param start_kernel (int), smallest kernel size
        :param end_karnel (int), largest kernel size
        :param kernel_size (int), number of opening operation to apply
        :param fetureLabel (list), list of features measure per cell (beside intensity), e.g. ["label","centroid",
            "area"] default ["label"]
        :return: dataframe of image intensity per segmented cell per image
        '''
        # returns column wise table
        tableInit, tableColumn = self.featuresTable(features =fetureLabel)
        open_vec = np.linspace(start_kernel, end_karnel, kernel_size, endpoint=True, dtype=int)
        if deploy:
            limit = 1
            open_vec = [open_vec[0],open_vec[-1]]
        else:
            limit = 3
        if len(open_vec) < limit:
            mesg = "increase kernel size"
            raise ValueError(mesg)
        for opning in open_vec:
            openImage = self.openingOperation(kernel = opning)
            tableGran_temp = self.measure_properties(input_image =openImage,input_mask = self.mask ,prop_names = fetureLabel)
            tableColumn[opning] = tableGran_temp['mean_int'].tolist()
        table_ = pd.DataFrame({'label':  tableInit.index.values})
        # convert pd to numpy preform normalisation by dividing
        kernel_array = [col for col in tableColumn.columns]
        mat = tableColumn.to_numpy(copy=True)
        imageSignalZero = mat[:, 0]
        tableColumnMatNorm = mat/imageSignalZero[:, np.newaxis]
        tableColumnNorm = pd.DataFrame(tableColumnMatNorm, columns=kernel_array)
        # adding label column
        table_gran_comp = pd.concat([table_, tableColumnNorm],axis = 1)
        table_gran_comp = table_gran_comp.melt(id_vars=["label"])
        tableFinal = table_gran_comp.sort_values(['label']).reset_index(drop=True)
        tableFinal = tableFinal.rename(
            columns={"label": "labeledCell", "variable":"kernel", "value": "intensity"})
        # cellLabel intoencoder
        tableFinal['labeledCell'] = pd.Categorical(tableFinal['labeledCell'], ordered=False)
        return tableFinal

    def stackTOobjects(self, extract_pixel, resize_pixel, img_label):
        '''
        fnction similar to the EBimage stackObjectsta, return a crop size based on center of measured mask
        :param: extract_pixel (int), size of extraction acording to mask (e.g. 50 pixel)
        :param: resize_pixel (int), resize for preforming tf prediction (e.g. 150 pixel)
        :param: img_label (img), the mask value for stack
        :return: (img), center image with out background
        '''
        img = self.image
        mask = self.mask
        table = pd.DataFrame(self.measure_properties(input_image = img, input_mask = mask ,prop_names = ["centroid"]))
        # table = table.astype({"centroid-0": 'int', "centroid-1": 'int'})
        x, y = table.loc[img_label, ["centroid-0", "centroid-1"]]
        x, y = int(x), int(y)
        mask_value = mask[x, y]
        x_start = x - extract_pixel
        x_end = x + extract_pixel
        y_start = y - extract_pixel
        y_end = y + extract_pixel
        if x_start < 0 or x_end < 0 or y_start < 0 or y_end < 0:
            return None, None
        else:
            mask_bin = np.zeros((np.shape(img)[0], np.shape(img)[1]), np.int32)
            mask_bin[mask == mask_value] = 1
            masked_image = img * mask_bin
            stack_img = masked_image[x_start:x_end, y_start:y_end]
            stack_img = resize(stack_img, (resize_pixel, resize_pixel), anti_aliasing=False)
            return stack_img, mask_value


class MERGE:
    '''
    merge granularity table
    This function is working with the glob function where the complete path is given
    '''
    @staticmethod
    def replaceLabel(arr):
        '''
        :param: arr (list), pandas series
        :return: replace label values with unique values
        '''
        uniqueLabel = np.unique(arr)
        idx = 0
        while idx <= len(uniqueLabel)-1:
            arr[arr==uniqueLabel[idx]] = randint(0, 500_000)
            idx += 1
        return arr

    def mergeTable(self,tableInput_name_list):
        '''
        :param: tableInput_name_list (list), array of csv file names (shuffled and merged)
        :return: merge all the tables
        '''
        idx = 0
        while idx <= len(tableInput_name_list)-1:
            if idx == 0:
                df = pd.read_csv(tableInput_name_list[0])
                idx += 1
                df['labeledCell'] = self.replaceLabel(df.labeledCell.values)
                imagename = tableInput_name_list[0].split("\\")[-1]
                df['Unnamed: 0'] = imagename
            else:
                df_temp = pd.read_csv(tableInput_name_list[idx])
                df_temp['labeledCell'] = self.replaceLabel(df_temp.labeledCell.values)
                imagename = tableInput_name_list[idx].split("\\")[-1]
                df_temp['Unnamed: 0'] = imagename
                df = pd.concat((df,df_temp),ignore_index=True,axis=0)
                idx += 1
        df = df.rename(columns={'Unnamed: 0':'image_name'})
        df = df.sort_values(by=['labeledCell', 'kernel'], na_position='first')
        return df

    @staticmethod
    def meanIntensity(df, group):
        '''
        :param: df (data frame) pandas Data frame
        :param: group (int) select the labeled Group
        :return: intensity percentage list and kernels
        '''
        dfSel = df.loc[df['classLabel'] == group]
        uniqueKernel = np.unique(dfSel.kernel.values)
        meanInten = []
        kernelArray = []
        for i, kernel in enumerate(uniqueKernel):
            arrTemp = dfSel.loc[dfSel['kernel'] == kernel, 'intensity'].to_list()
            # clean array
            arrTempClean = [arr for arr in arrTemp if str(arr) != 'nan']
            meanInten.append(sum(arrTempClean) / len(arrTemp))
            kernelArray.append(kernel)
        return meanInten, kernelArray

    @staticmethod
    def cleanTable(df):
        '''
        :param: df (data frame), pandas Data frame
        :param: group (int), selected labeled Group
        :return: remove null observations
        '''
        dfClean = df.copy()
        uniqeImageLabel = np.unique(df.labeledCell.values)
        for imagename in uniqeImageLabel:
            df_temp = df.loc[df['labeledCell']==imagename,:]
            arr = df_temp.intensity.values.tolist()
            arr.pop(0)
            arrTemp = np.mean(arr)
            if not(arrTemp>0):
                idx = dfClean.loc[df['labeledCell']==imagename,:].index
                dfClean.drop(idx, inplace=True)
        return dfClean



    @staticmethod
    def calcDecay(df,kernelSelect):
        '''
        :param: df (data frame), pandas finel table
        :param: kernelSelect (int) kernel size selected
        :return: filter table for the kernel selcted result in 50% reduction in signal
        '''
        return df.loc[df['kernel']==kernelSelect,:]








import re
import numpy as np
import pandas as pd
from sklearn import preprocessing
from tqdm import tqdm


class mapSgRNA():
    def __init__(self,df1,df2):
        '''
        :param df1: Q1 sample (original)
        :param df2: Q2 sample (active sort)
        '''
        self.df1 = df1
        self.df2 = df2
        # adding 2 to all the data for preforming log transformation
        self.df1[self.df1.columns[2]] = self.df1[self.df1.columns[2]] + 2
        self.df2[self.df1.columns[2]] = self.df2[self.df2.columns[2]] + 2
    def mapping(self, uniqueSgRNA):
        '''
        :param uniqueSgRNA: list of strings, unique sgRNA
        :return: mapping reads according to unique sgRNA
        '''
        sgRNAdic = {'Gene':[],'sgRNA':[],"reads_ctrl":[],"reads_activate":[]}
        for sgRNA in tqdm(uniqueSgRNA):
            sgRNAdic['sgRNA'].append(sgRNA)
            # check for non targeting:
            if re.match("non-targeting.*", sgRNA) is None:
                sgRNAdic['Gene'].append(re.sub('_._.*','', sgRNA))
            else:
                sgRNAdic['Gene'].append("non")
            # reads ctrl
            if self.df1.loc[self.df1['sgID'] == sgRNA, self.df1.columns[2]].any():
                sgRNAdic['reads_ctrl'].append(self.df1.loc[self.df1['sgID'] == sgRNA, self.df1.columns[2]].values[0])
            else:
                sgRNAdic['reads_ctrl'].append(2)
            # reads active
            sgRNAdic['reads_activate'].append(self.df2.loc[self.df2['sgID'] == sgRNA, self.df2.columns[2]].values[0])
        return sgRNAdic

    @staticmethod
    def dataFrameFinal(sgRNAdic):
        sgRNAdic['log2FoldChange'] = np.log(sgRNAdic['reads_activate']) - np.log(sgRNAdic['reads_ctrl'])
        sgRNAdic['scaleLog2FoldChange'] = preprocessing.scale(sgRNAdic['log2FoldChange'])
        df = pd.DataFrame(sgRNAdic)
        return df




import numpy as np
from skimage.exposure import rescale_intensity, histogram
import matplotlib.pyplot as plt
import matplotlib as mpl

import plotly.express as px
from skimage import io, filters, measure, color, img_as_ubyte

from AIPyS.supportFunctions import display_and_xml as dx


def image_with_contour(img, active_labels, data_table, active_columns, color_column):
    """
    Returns a greyscale image that is segmented and superimposed with contour traces of
    the segmented regions, color coded by values from a data table.
    Parameters
    ----------
    img : PIL Image object.
    active_labels : list
        the currently visible labels in the datatable
    data_table : pandas.DataFrame
        the currently visible entries of the datatable
    active_columns: list
        the currently selected columns of the datatable
    color_column: str
        name of the datatable column that is used to define the colorscale of the overlay
    """

    # First we get the values from the selected datatable column and use them to define a colormap
    values = np.array(data_table[color_column].values)
    norm = mpl.colors.Normalize(vmin=values.min(), vmax=values.max())
    cmap = plt.cm.get_cmap("plasma")

    # Now we convert our background image to a greyscale bytestring that is very small and can be transferred very
    # efficiently over the network. We do not want any hover-information for this image, so we disable it
    fig = px.imshow(img, binary_string=True, binary_backend="jpg",)
    fig.update_traces(hoverinfo="skip", hovertemplate=None)

    # For each region that is visible in the datatable, we compute and draw the filled contour, color it based on
    # the color_column value of this region, and add it to the figure
    # here is an small tutorial of this: https://scikit-image.org/docs/dev/auto_examples/segmentation/plot_regionprops.html#sphx-glr-auto-examples-segmentation-plot-regionprops-py
    for rid, row in data_table.iterrows():
        label = row.label
        value = row[color_column]
        contour = measure.find_contours(active_labels == label, 0.5)[0]
        # We need to move the contour left and up by one, because
        # we padded the label array
        y, x = contour.T - 1
        # We add the values of the selected datatable columns to the hover information of the current region
        hoverinfo = (
            "<br>".join(
                [
                    # All numbers are passed as floats. If there are no decimals, cast to int for visibility
                    f"{prop_name}: {f'{int(prop_val):d}' if prop_val.is_integer() else f'{prop_val:.3f}'}"
                    if np.issubdtype(type(prop_val), "float")
                    else f"{prop_name}: {prop_val}"
                    for prop_name, prop_val in row[active_columns].iteritems()
                ]
            )
            # remove the trace name. See e.g. https://plotly.com/python/reference/#scatter-hovertemplate
            + " <extra></extra>"
        )
        fig.add_scatter(
            x=x,
            y=y,
            name=label,
            opacity=0.2,
            mode="lines",
            line=dict(color=mpl.colors.rgb2hex(cmap(norm(value))),),
            fill="toself",
            customdata=[label] * len(x),
            showlegend=False,
            hovertemplate=hoverinfo,
            hoveron="points+fills",
        )

    # Finally, because we color our contour traces one by one, we need to manually add a colorscale to explain the
    # mapping of our color_column values to the colormap. This also gets added to the figure
    fig.add_scatter(
        # We only care about the colorscale here, so the x and y values can be empty
        x=[None],
        y=[None],
        mode="markers",
        showlegend=False,
        marker=dict(
            colorscale=[mpl.colors.rgb2hex(cmap(i)) for i in np.linspace(0, 1, 50)],
            showscale=True,
            # The cmin and cmax values here are arbitrary, we just set them to put our value ticks in the right place
            cmin=-5,
            cmax=5,
            colorbar=dict(
                tickvals=[-5, 5],
                ticktext=[f"{np.min(values[values!=0]):.2f}", f"{np.max(values):.2f}",],
                # We want our colorbar to scale with the image when it is resized, so we set them to
                # be a fraction of the total image container
                lenmode="fraction",
                len=0.6,
                thicknessmode="fraction",
                thickness=0.05,
                outlinewidth=1,
                # And finally we give the colorbar a title so the user may know what value the colormap is based on
                title=dict(text=f"<b>{color_column.capitalize()}</b>"),
            ),
        ),
        hoverinfo="none",
    )

    # Remove axis ticks and labels and have the image fill the container
    fig.update_layout(margin=dict(l=0, r=0, b=0, t=0, pad=0), template="simple_white")
    fig.update_xaxes(visible=False, range=[0, img.width]).update_yaxes(
        visible=False, range=[img.height, 0]
    )
    return fig
#https://dash.plotly.com/datatable/conditional-formatting
def row_highlight(roi_list_ctrl,roi_list_target):
    '''
        Componant of Dash datatable - highlight raws in the table
        :parameter
        roi_list_ctrl - list of ROI - in red #F31515
        roi_list_target -  list of ROI - in green #1ABA19
    '''
    return  ([
                 {'if': {'filter_query': '{{label}} = {}'.format(int(roi_ctrl))},
                     'backgroundColor': '{}'.format('#F31515'),
                     'color': 'white'
                 }
                for roi_ctrl in roi_list_ctrl
                ] +
                [
                 {
                     'if':
                         {'filter_query': '{{label}} = {}'.format(int(roi_))},
                        'backgroundColor': '{}'.format('#1ABA19'),
                        'color': 'white'
                 }
                for roi_ in roi_list_target
            ])

def countor_map(mask_target,roi_ctrl,roi_target,ch2_rgb):
    ''':parameter
        mask_target - contour target channel
        ROI - current click point and the list of the last clicks
        ch2_rgb - with seed is displayed in blue
        return:
        an rGB image with seed and clicked target segment map.
    '''
    if len(roi_ctrl) > 0:
        bf_mask_sel_ctrl = np.zeros(np.shape(mask_target), dtype=np.int32)
        #adding ctrl map
        for list in roi_ctrl:
            bf_mask_sel_ctrl[mask_target == list] = list
        c_mask_ctrl = dx.binary_frame_mask(ch2_rgb, bf_mask_sel_ctrl)
    else:
        c_mask_ctrl = np.zeros(np.shape(mask_target), dtype=np.int32)
    # adding target map
    if len(roi_target) > 0:
        bf_mask_sel_trgt = np.zeros(np.shape(mask_target), dtype=np.int32)
        for list in roi_target:
            bf_mask_sel_trgt[mask_target == list] = list
        c_mask_trgt = dx.binary_frame_mask(ch2_rgb, bf_mask_sel_trgt)
    else:
        c_mask_trgt = np.zeros(np.shape(mask_target), dtype=np.int32)
    ch2_rgb[c_mask_ctrl > 0, 0] = 255
    ch2_rgb[c_mask_trgt > 0, 1] = 255
    return ch2_rgb






from AIPyS.classification.bayes.GranularityDataGen import GranularityDataGen
from AIPyS.supportFunctions.GranularityFunc import openingOperation
import numpy as np
import pandas
import string
import cv2
import pdb
import random
import skimage
import os
import pandas as pd
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
from IPython.display import clear_output

class GranularityDeploy(GranularityDataGen):
    '''
    This function is designed to analyze an image by employing a sliding window approach. As input, it accepts an image and systematically traverses it with a running window. For each segment captured within this window, the function utilizes a Bayesian model to predict the class to which the segment belongs. Upon completing the traversal, it generates a comprehensive table. This table showcases the predictions for each slice, alongside the precise coordinates. These coordinates are instrumental in selecting and referencing the relevant entries in the class prediction table, thereby facilitating a detailed and organized analysis of the image based on the Bayesian classification.
    '''
    def __init__(self,a,b,td,expansion_factor,*args, **kwargs):
        self.a = a #intercept
        self.b = b #slope
        self.td = td
        self.expansion_factor = expansion_factor
        super().__init__(*args, **kwargs)

    def classify(self,Int):
        mu = self.a + self.b * Int
        prob = 1 / (1 + np.exp(-mu))
        return prob

    def creatCrudeMask(self,img_curr,x_start,y_start,winSize):
        height, width = img_curr.shape[:2]
        center_x, center_y = x_start + winSize // 2, y_start + winSize // 2
        new_winSize = int(winSize * self.expansion_factor)
        center_x, center_y = x_start + winSize // 2, y_start + winSize // 2
         # Adjusting new window coordinates based on the image boundaries
        new_x_start = max(center_x - new_winSize // 2, 0)
        new_y_start = max(center_y - new_winSize // 2, 0)
        new_x_end = min(center_x + new_winSize // 2, width)
        new_y_end = min(center_y + new_winSize // 2, height)
    
        # Handling cases where the window hits the boundary and keeping the aspect ratio
        if new_x_end - new_x_start < new_winSize:
            diff = new_winSize - (new_x_end - new_x_start)
            new_x_start = max(new_x_start - diff, 0)
            new_x_end = min(new_x_end + diff, width)
        if new_y_end - new_y_start < new_winSize:
            diff = new_winSize - (new_y_end - new_y_start)
            new_y_start = max(new_y_start - diff, 0)
            new_y_end = min(new_y_end + diff, height)
        new_x_start = max(0, new_x_start)
        new_y_start = max(0, new_y_start)
        new_x_end = min(new_x_end, width)
        new_y_end = min(new_y_end, height)
        return (new_x_start, new_y_start, new_x_end, new_y_end)

    
    def predictCrop(self,savedict = False):
        img = self.img # should be single image
        mask = self.cytosolSegmentation(img)
        img_crudeMasked = np.zeros_like(img) # container for masked image
        dataContainer = {'x':[],'y':[],'ratio':[],'pred':[]}
        if not isinstance(self.windowSize, (int, float)):
                raise ValueError("window size is missing")
        if not isinstance(self.rnadomWindows, (int, float)):
            raise ValueError("rando crop generator is missing choose a number")
        minAxsis = min(img.shape[0],img.shape[1])
        winSize = self.windowSize
        while minAxsis % winSize != 0:
            minAxsis -= 1
        stepSize =  int(minAxsis/winSize)
        print(f'max image size: {minAxsis}, step size: {stepSize}')
        for x in range(0,minAxsis,stepSize):
            for y in range(0,minAxsis,stepSize):
                imageCrop = img[x:x+stepSize,y:y+stepSize]
                maskCrop = mask[x:x+stepSize,y:y+stepSize]
                imageDecy = openingOperation(kernel = self.kernelGran, image = imageCrop)
                if np.sum(imageDecy[maskCrop>0]) == 0:
                    denominator = np.sum(imageCrop[maskCrop > 0]) + 0.000001
                else:
                    denominator = np.sum(imageCrop[maskCrop > 0])
                ratio = np.sum(imageCrop[maskCrop > 0] - imageDecy[maskCrop > 0])/denominator
                pred = self.classify(ratio)
                if savedict:
                    dataContainer['ratio'].append(ratio)
                    dataContainer['pred'].append(pred)
                    dataContainer['x'].append(x)
                    dataContainer['y'].append(y)
                else:
                    # preform segmentation on the fly
                    if pred > self.td:
                        xs,ys,xe,ye = self.creatCrudeMask(imageCrop,x,y,winSize = stepSize)
                        img_crudeMasked[xs:xe,ys:ye] = img[xs:xe,ys:ye] #insert selected cell 
        if savedict:
            return dataContainer
        else:
            return img_crudeMasked


    


import h5py
import numpy as np
import os
import shutil
from pathlib import Path
import argparse

# Define your parameters
params = {
    "data_dir": None,
    "files": None,
    "videoName": None,
    "Image_name": None,
    "outPath": None,
    "block_size_cyto": 3,
    "offset_cyto": -5,
    "global_ther": 0.51,
    "clean": 3,
    "channel": 0,
    "bit8": 20,
    "ci": 3,
    "start_kernel": 2,
    "end_karnel": 50, 
    "kernel_size": 20,
    "outputImageSize": 500,
    "windowSize": 6,
    "rnadomWindows": 10,
    "kernelGran": 4,
    "w": 500,
    "h": 500,
    "diameter":50,
    "model_type":'cyto',
    "channels":'grayscale',
    "extract_pixel":50,
    "resize_pixel":150,
    "dataPath":None,
    "imagePath":None,
    "imW":10,
    "imH":10,
    "thold":0.5,
    "imagesN":30,
    "trainingDataPath":None,
    "areaSel":-1,
    "fractionData":-1,
    "intercept":0.0,
    "slope":0.0
    }

# Specify the HDF5 file name
user_parameters_path = os.path.join(Path.home(), '.AIPyS', 'parameters.h5')
dirPath = os.path.join(Path.home(), '.AIPyS') 

def parametersGnert():
    if not os.path.exists(dirPath):
       os.makedirs(os.path.dirname(dirPath), exist_ok=True) 
    # Open a new HDF5 file
    with h5py.File(user_parameters_path, 'w') as hdf:
        # Iterate through dictionary items and save them to the HDF5 file
        for key, value in params.items():
            # Convert None values to a recognizable string, as NoneType can't be directly stored
            if value is None:
                value = 'None'
            # Check if value needs to be stored as a string
            if isinstance(value, str):
                dt = h5py.special_dtype(vlen=str)  # special dtype for variable-length strings
                dset = hdf.create_dataset(key, (1,), dtype=dt)
                dset[0] = value
            else:
                # Directly store if the value is numeric
                hdf.create_dataset(key, data=value)
    print(f"Parameters saved to {user_parameters_path}")

def resetParameters():
    parametersGnert()
    print(f'Done')

def display_parameters():
    """Read and display parameters from the user's parameters.h5 file."""
    with h5py.File(user_parameters_path, 'r') as hdf:
        print("Current parameters stored in the H5 file:")
        for key in hdf.keys():
            print(f"{key}: {hdf[key][()]}")

def main():
    parser = argparse.ArgumentParser(description='Manage parameters file', formatter_class=argparse.RawTextHelpFormatter)
    parser.add_argument('action', choices=['generate', 'reset', 'display'],
                        help="generate: Create a new parameters file with default values."
                             "reset: Reset the parameters file to default values."
                             "display: Show current values in the parameters file.")
    args = parser.parse_args()
    
    if args.action == 'generate':
        parametersGnert()
    elif args.action == 'reset':
        resetParameters()
    elif args.action == 'display':
        display_parameters()
    else:
        raise ValueError("Unknown action. Available actions: generate, reset, display")

if __name__ == "__main__":
    main()


import argparse
import random
import glob
from pathlib import Path
import os
from AIPyS.segmentation.cellpose.plotObjectAreas import plotObjectAreas
from AIPyS.supportFunctions.unpack_h5 import parametersInspec
from web_app.AreasViz.app import area_viz

def main():
    parameters = parametersInspec(option,user_parameters_path)
    if parameters['channels']=='greyscale':
        channels = [0,0]
    else:
        print('channels are missing')
    pattern = os.path.join(parameters['trainingDataPath'], '**', '*.tif')
    # List all the .tif files
    files = glob.glob(pattern, recursive=True)
    random.shuffle(files)
    plotObjectAreas(kernelGran = parameters['kernelGran'],w = parameters['w'],h = parameters['h'],
                        extract_pixel = parameters['extract_pixel'], resize_pixel = parameters['resize_pixel'],diameter =  parameters['diameter'],
                        model_type = parameters['model_type'], channels = channels,
                        Image_name = files[:parameters['imagesN']], outPath = parameters['outPath'],trainingDataPath = parameters['trainingDataPath'])
    print("Web application initiate")
    dataPath = parameters['dataPath']
    imagePath = parameters['imagePath']
    app = area_viz(imagePath,dataPath)


from AIPyS.segmentation.parametric.GlobalSeg import GlobalSeg
import numpy as np

class RunningWindow(GlobalSeg):
    '''
    Split the image according to the window size to identify the smallest image size with the highest possibility of identifying the phenotypic cell. Expect single image
    '''
    def __init__(self,windowSize,rnadomWindows,*args, **kwargs):
        self.windowSize = windowSize # for example split image to 6 slides
        self.rnadomWindows = rnadomWindows # for example add 3 more random slides
        super().__init__(*args, **kwargs)
        
    def runningWin(self,image_input,mask):
        '''
        creates a running window sized by splitting the image to equal size images
        '''
        if not isinstance(self.windowSize, (int, float)):
                raise ValueError("window size is missing")
        if not isinstance(self.rnadomWindows, (int, float)):
            raise ValueError("rando crop generator is missing choose a number")
        minAxsis = min(image_input.shape[0],image_input.shape[1])
        winSize = self.windowSize
        while minAxsis % winSize != 0:
            minAxsis -= 1
        stepSize =  int(minAxsis/winSize)
        print(f'max image size: {minAxsis}, step size: {stepSize}')
        imageSlice = []
        maskSlice = []
        for i in range(0,minAxsis,stepSize):
            for j in range(0,minAxsis,stepSize):
                imageSlice.append(image_input[i:i+stepSize,j:j+stepSize])
                maskSlice.append(mask[i:i+stepSize,j:j+stepSize])
        # adding rando slices
        rnadomWindows = [1]*self.rnadomWindows
        range_xy = 0+stepSize,minAxsis-250
        while rnadomWindows:
            x = np.random.randint(low = range_xy[0],high = range_xy[1],size =1).tolist()[0]
            imageSlice.append(image_input[x:x+stepSize,x:x+stepSize])    
            maskSlice.append(mask[x:x+stepSize,x:x+stepSize])
            rnadomWindows.pop()
        return imageSlice, maskSlice



import xml.etree.ElementTree as xml
import matplotlib.pyplot as plt
import numpy as np
from PIL import Image,ImageEnhance
# from PIL import fromarray
import plotly.express as px
from skimage.filters import threshold_local
from scipy.ndimage.morphology import binary_opening
from skimage import io, filters, measure, color, img_as_ubyte
import skimage.morphology as sm
from skimage.segmentation import watershed
from skimage import measure
from skimage.exposure import rescale_intensity
from sklearn.mixture import GaussianMixture
import os
import pandas as pd
from scipy.ndimage.morphology import binary_fill_holes
import base64
from datetime import datetime
import re
from AIPyS.supportFunctions.display_and_xml import unique_rand
from AIPyS.supportFunctions import display_and_xml as dx


def image_to_8bits(input):
    imType = str(input.dtype)
    if not ('uint8' in imType):
        input = img_as_ubyte(input)
    if len(np.shape(input)) < 3:
        input = input.reshape(1, np.shape(input)[0], np.shape(input)[1])
    img = input
    return img

def show_image_adjust(image, low_prec, up_prec):
    """
    image= np array 2d
    low/up precentile border of the image
    """
    percentiles = np.percentile(image, (low_prec, up_prec))
    scaled_ch1 = rescale_intensity(image, in_range=tuple(percentiles))
    return scaled_ch1
    # PIL_scaled_ch1 = Image.fromarray(np.uint16(scaled_ch1))
    # PIL_scaled_ch1.show()
    # return PIL_scaled_ch1

def px_pil_figure(img,bit,mask_name,fig_title,wh,*args):
    '''
    :param img: image input - 3 channel 8 bit image
             bit:1 np.unit16 or 2 np.unit8
             fig_title: title for display on dash
             wh: width and hight in pixels
    :return: encoded_image (e_img)
    '''
    bit = str(img.dtype)
    if bit == "bool":
        # binary
        im_pil = Image.fromarray(img)
    elif bit == "int64":
        # 16 image (normal image)
        im_pil = Image.fromarray(np.uint16(img))
    else:
        # ROI mask
        img_gs = img_as_ubyte(img)
        im_pil = Image.fromarray(img_gs)
        if len(args) > 0:
            enhancer = ImageEnhance.Contrast(im_pil)
            factor = args[0]  # gives original image
            im_pil = enhancer.enhance(factor)
    fig_ch = px.imshow(im_pil, binary_string=True, binary_backend="jpg", width=wh, height=wh,title=fig_title,binary_compression_level=9).update_xaxes(showticklabels=False).update_yaxes(showticklabels = False)
    fig_ch.update_layout(title_x=0.5)
    return fig_ch


def XML_creat(filename,block_size,offset,rmv_object_nuc,block_size_cyto,offset_cyto,global_ther,rmv_object_cyto,rmv_object_cyto_small):
    root = xml.Element("Segment")
    cl = xml.Element("segment") #chiled
    root.append(cl)
    block_size_ = xml.SubElement(cl,"block_size")
    block_size_.text = block_size
    offset_ = xml.SubElement(cl,"offset")
    offset_.text = "13"
    rmv_object_nuc_ = xml.SubElement(cl, "rmv_object_nuc")
    rmv_object_nuc_.text = "rmv_object_nuc"
    block_size_cyto_ = xml.SubElement(cl, "block_size_cyto")
    block_size_cyto_.text = "block_size_cyto"
    offset_cyto_ = xml.SubElement(cl, "offset_cyto")
    offset_cyto_.text = "offset_cyto"
    global_ther_ = xml.SubElement(cl, "global_ther")
    global_ther_.text = "global_ther"
    rmv_object_cyto_ = xml.SubElement(cl, "rmv_object_cyto")
    rmv_object_cyto_.text = "rmv_object_cyto"
    rmv_object_cyto_small_ = xml.SubElement(cl, "rmv_object_cyto_small")
    rmv_object_cyto_small_.text = "rmv_object_cyto_small"
    tree = xml.ElementTree(root)
    with open(filename,'wb') as f:
        tree.write(f)

def seq(start, end, by=None, length_out=None):
    len_provided = True if (length_out is not None) else False
    by_provided = True if (by is not None) else False
    if (not by_provided) & (not len_provided):
        raise ValueError('At least by or n_points must be provided')
    width = end - start
    eps = pow(10.0, -14)
    if by_provided:
        if (abs(by) < eps):
            raise ValueError('by must be non-zero.')
        # Switch direction in case in start and end seems to have been switched (use sign of by to decide this behaviour)
        if start > end and by > 0:
            e = start
            start = end
            end = e
        elif start < end and by < 0:
            e = end
            end = start
            start = e
        absby = abs(by)
        if absby - width < eps:
            length_out = int(width / absby)
        else:
            # by is too great, we assume by is actually length_out
            length_out = int(by)
            by = width / (by - 1)
    else:
        length_out = int(length_out)
        by = width / (length_out - 1)
    out = [float(start)] * length_out
    for i in range(1, length_out):
        out[i] += by * i
    if abs(start + by * length_out - end) < eps:
        out.append(end)
    return out

def rgb_file_gray_scale(input_gs_image,mask=None,channel=None):
    ''' create a 3 channel rgb image from 16bit input image
        optional bin countor image from ROI image
        :parameter
        input_gs_image: 16bit nparray
        mask: 32int roi image
        channel: 0,1,2 (rgb)
        :return
        3 channel stack file 8bit image
    '''
    input_gs_image = (input_gs_image / input_gs_image.max()) * 255
    ch2_u8 = np.uint8(input_gs_image)
    rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
    rgb_input_img[:, :, 0] = ch2_u8
    rgb_input_img[:, :, 1] = ch2_u8
    rgb_input_img[:, :, 2] = ch2_u8
    if mask is not None and len(np.unique(mask)) > 1:
        bin_mask = dx.binary_frame_mask(ch2_u8, mask)
        bin_mask = np.where(bin_mask == 1, True, False)
        if channel is not None:
            rgb_input_img[bin_mask > 0, channel] = 255
        else:
            rgb_input_img[bin_mask > 0, 2] = 255
    return rgb_input_img

def rgbTograyscale(input):
    '''
        create a 3 channel grayscale from np format (W,H,C)
        :parameter
        input_gs_image: 8 bit
        :return
        3 channel stack file 8bit image
    '''
    rgb_input_img = np.zeros((np.shape(input)[2], np.shape(input)[0], np.shape(input)[1]), dtype=np.uint8)
    for i in range(np.shape(input)[2]):
        rgb_input_img[i, :, :] = input[:, :, i]
    return rgb_input_img


def gray_scale_3ch(input_gs_image):
    input_gs_image = (input_gs_image / input_gs_image.max()) * 255
    ch2_u8 = np.uint8(input_gs_image)
    rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
    rgb_input_img[:, :, 0] = ch2_u8
    rgb_input_img[:, :, 1] = ch2_u8
    rgb_input_img[:, :, 2] = ch2_u8
    return rgb_input_img

def plot_composite_image(img,mask,fig_title,alpha=0.2):
    # apply colors to mask
    '''
    :param img: input 3 channel grayscale image
    :param mask: mask
    :param fig_title: title
    :param alpha: transprancy for blending
    :param img_shape:
    :return:
    '''
    mask = np.array(mask, dtype=np.int32)
    mask_deci = (mask - np.min(mask)) / (np.max(mask) - np.min(mask))
    cm = plt.get_cmap('CMRmap')
    colored_image = cm(mask_deci)
    colored_image = (colored_image[:, :, :3] * 255).astype(np.uint8)
    #RGB pil image
    img_mask = img_as_ubyte(colored_image)
    im_mask_pil = Image.fromarray(img_mask).convert('RGB')
    img_gs = img_as_ubyte(img)
    im_pil = Image.fromarray(img_gs).convert('RGB')
    im3 = Image.blend(im_pil, im_mask_pil, alpha)
    fig_ch = px.imshow(im3, binary_string=True, binary_backend="jpg", width=700, height=700, title=fig_title,
                       binary_compression_level=0).update_xaxes(showticklabels=False).update_yaxes(showticklabels=False)
    fig_ch.update_layout(title_x=0.5)
    return fig_ch

def save_pil_to_directory(img,bit,mask_name,output_dir = 'temp',mask = None, merge_mask = None,channel=None):
    '''
    Save image composite with ROI
    parameters
    ----------
    image input
    bit:1 
        np.unit16 or 2 np.unit8
    merge_mask: 
        ROI - ROI + 3ch greyscale input OR  -BIN -  bin + 3ch greyscale
    channel
        for display Bin merge  with rgb input
    return 
        encoded_image (e_img)
    '''
    bit = str(img.dtype)
    if merge_mask=='ROI':
        # img  must be 3 channel grayscale
        mask = np.array(mask, dtype=np.int32)
        mask_deci = (mask - np.min(mask)) / (np.max(mask) - np.min(mask))
        cm = plt.get_cmap('CMRmap')
        colored_image = cm(mask_deci)
        colored_image = (colored_image[:, :, :3] * 255).astype(np.uint8)
        # RGB pil image
        img_mask = img_as_ubyte(colored_image)
        im_mask_pil = Image.fromarray(img_mask).convert('RGB')
        img_gs = img_as_ubyte(img)
        im3 = Image.fromarray(img_gs).convert('RGB')
        im_pil = Image.blend(im3, im_mask_pil, 0.2)
    elif merge_mask=='BIN':
        #img  must be 3 channel grayscale
        img_bin = rgb_file_gray_scale(img, mask=mask, channel=channel)
        img_gs = img_as_ubyte(img_bin)
        im_pil = Image.fromarray(img_gs)
    elif merge_mask is None:
        if bit == "bool":
            # binary
            im_pil = Image.fromarray(img)
        elif bit == "int64":
            # 16 image (normal image)
            im_pil = Image.fromarray(np.uint16(img))
        else:
            # ROI mask
            img = np.uint8(img)
            roi_index_uni = np.unique(img)
            roi_index_uni = roi_index_uni[roi_index_uni > 1]
            sort_mask_buffer = np.ones((np.shape(img)[0], np.shape(img)[1], 3), dtype=np.uint8)
            for npun in roi_index_uni:
                for i in range(3):
                    sort_mask_buffer[img == npun, i] = unique_rand(2, 255, 1)[0]
            im_pil = Image.fromarray(sort_mask_buffer, mode='RGB')
    filename1 = datetime.now().strftime("%Y%m%d_%H%M%S" + mask_name)
    im_pil.save(os.path.join(output_dir, filename1 + ".png"), format='png')  # this is for image processing
    e_img = base64.b64encode(open(os.path.join('temp', filename1 + ".png"), 'rb').read())
    return e_img

def remove_gradiant_label_border(mask):
    '''
    mask from rescale return with border gradiant which is for the borders
    :return mask no borders
    '''
    mask_ = np.where(mask > 0, mask, 0)
    mask_intact = np.where(np.mod(mask_, 1) > 0, 0, mask_)
    mask_intact = np.array(mask_intact, np.uint32)
    return mask_intact

def deployRatioCalc(image_input,mask):
    return pd.DataFrame(measure.regionprops_table(mask,intensity_image=image_input,properties=['area', 'label', 'centroid','intensity_mean']))

def call_bin(table_sel_cor, threshold ,img_blank):
    '''
    :parameter:
    table_sel_cor: pandas table contain the center coordinates
    threshold: thershold for predict phenotype (e.g. 0.5)
    img_blank: blank image in the shape of the input image
    :return: binary image of the called masks, table_sel
    '''
    table_na_rmv_trgt = table_sel_cor.loc[table_sel_cor['predict'] > threshold, :]
    for label in table_na_rmv_trgt.index.values:
        x, y = table_na_rmv_trgt.loc[label,["centroid-0", "centroid-1"]].tolist()
        row, col = disk((int(x), int(y)), 10)
        img_blank[row, col] = 1
    return img_blank, table_na_rmv_trgt
    
def keepObject(table,mask):
    '''
    :param table: keep all the object which are predicted above the threshold
    :return: ROI image mask of selected objects
    '''
    nmask = np.zeros_like(mask)
    for label in table.index.values:
        nmask[mask == label] = label
    return nmask

def areaThreshold(arr_area):
    y = np.array(arr_area).reshape(-1, 1)  # Reshape data to 2D array required by GMM
    # Define the range of components to test
    n_components = np.arange(1, 3)
    # Fit GMM models for 1 and 2 components
    models = [GaussianMixture(n_components=n, random_state=42).fit(y) for n in n_components]
    # Evaluate models using Bayesian Information Criterion (BIC)
    BIC_scores = [model.bic(y) for model in models]
    # Select the model with the lowest BIC
    best_model = models[np.argmin(BIC_scores)]

    # Assuming the best model has two components (as expected)
    if best_model.n_components == 2:
        means = np.sort(best_model.means_.flatten())
        std_dev = np.sqrt(best_model.covariances_.flatten())
            # Estimate a potential threshold
        try:
            threshold = (means[0] + means[1]) / 2
        except Exception:
            threshold = -1
        if isinstance(threshold, (int, float)) and threshold > 0:
            return threshold
        else:
            threshold =  "na"
            return threshold
    threshold =  "na"
    return threshold
    


from setuptools import setup, find_packages

with open('requirements.txt') as f:
    required = f.read().splitlines()
setup(
    name="AIPyS",
    version="0.1.3",
    packages=find_packages(include=['AIPyS','AIPyS.CLI','AIPyS.classification.bayes','AIPyS.classification.CNN',
                                    'AIPyS.segmentation.cellpose','AIPyS.segmentation.parametric','AIPyS.supportFunctions',
                                    'web_app.Image_labeling','web_app.measure_length','web_app.measure_length','web_app.TableViz', 'web_app.AreasViz',]),
    install_requires=required,
    entry_points={
        'console_scripts': [
           'aipys=AIPyS.CLI.aipys:main',  
           'updateParameters=AIPyS.CLI.set_parameters:main',
            'load-parameters=AIPyS.CLI.loadParameters:main', 
            ],
    },
    author="Gil Kanfer",
    author_email="gil.kanfer.il@gmail.com",
    description="AI Powered Photoswitchable Screen",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="",  # Consider adding a URL if you have a project page or source repository
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.8',  # Specify the minimum version of Python required
)


import numpy as np
import string
import cv2
import pdb
import random
import skimage
import os
import pandas as pd
import matplotlib.pyplot as plt
from sklearn.mixture import GaussianMixture
from PIL import Image, ImageEnhance, ImageDraw,ImageFont
from IPython.display import clear_output
from AIPyS.supportFunctions.AIPS_file_display import Compsite_display
from AIPyS.segmentation.cellpose.AIPS_cellpose import AIPS_cellpose
from AIPyS.supportFunctions.GranularityFunc import resize

class plotObjectAreas(StackObjects_cellpose):
    def __init__(self,trainingDataPath, w, h ,*args, **kwargs):
        # require self.outPath(DataLoad) 
        self.trainingDataPath = trainingDataPath
        self.w, self.h =w,h # for resizing
        super().__init__(*args, **kwargs)
        self.areaCalc_imageGen_cp()
    
    def generate_random_string(self,length):
        letters = string.ascii_letters  # this includes both lowercase and uppercase letters
        result_str = ''.join(random.choice(letters) for i in range(length))
        return result_str 

    def getMeanSD(self,image_input):
        return np.round(np.mean(image_input),3),np.round(np.std(image_input),3)
    
    def check_list(self):
        if not isinstance(self.Image_name, list):
            raise ValueError("Requries list of images path")
    
    def sdCalc(self,image,mean_value):
        squared_diff_sum = 0
        num_elements = 0
        for row in image:
            for element in row:
                squared_diff_sum += (element - mean_value) ** 2
                num_elements += 1
        return (squared_diff_sum / num_elements)**0.5
        
    
    def areaCalc_imageGen_cp(self):
        '''
        Adjusted to cellpose, Generates image per slice and a single table. for training
        Note - expects a list of images path
        Produce output table used for Bayes training (without the label option)
        and Video file with Ratio value and image file name.
        Ratio is the granular spectrum
        '''
        self.check_list()
        image_name_list = self.Image_name
        tempFolder = os.path.join(self.outPath,'AreaAnalysis')
        ImagePath = os.path.join(tempFolder,'images')
        DataPath = os.path.join(tempFolder,'data')
        os.makedirs(tempFolder, exist_ok=True)
        os.makedirs(ImagePath, exist_ok=True)
        os.makedirs(DataPath, exist_ok=True)
        dfGran = {'name':[],'intensity':[],'sd':[],'maskArea':[]}
        c = 0
        for image_path in image_name_list:
            image = skimage.io.imread(image_path)
            image_name = self.generate_random_string(7)
            # Initiate cellpose segmentation
            mask, table = self.cellpose_segmentation(image_input = image)
            # create original image:
            #print('-' * c, end = '\r')
            clear_output(wait=True)
            for idx, (index, row) in enumerate(table.iterrows()):
                if idx % 10 == 0:
                    c += 1
                    print('.'*c, end = '\r')
                img_label = row['label'] # first cell analysis 
                try:
                    stack_img, stack_mask = self.stackObjects_cellpose_ebimage_parametrs_method(image,mask,table,img_label)
                except:
                    continue
                if stack_img is None:
                    continue
                intensity =  np.mean(stack_img)
                maskArea = row['area']
                sd = self.sdCalc(stack_img,intensity)
                #image intensity and sd before opening:
                dfGran['name'].append(image_name + f'_{idx}.png')
                dfGran['intensity'].append(intensity)
                dfGran['sd'].append(sd)
                dfGran['maskArea'].append(maskArea)
                frame_obj = Compsite_display(stack_img, stack_mask, 2)
                frame = resize(image=frame_obj.rgb_out, width=self.w, hight=self.h)
                PIL_image_grey = Image.fromarray(frame).convert('L') # convery to gray scale
                PIL_image = PIL_image_grey.convert('RGB') # for getting a grey scale with red text
                draw = ImageDraw.Draw(PIL_image)
                font_size = 24
                font = ImageFont.truetype("arial.ttf", font_size)  # Adjust as necessary
                draw.text((5, 5),f'Imagename:{image_name}_{idx} \n Area: {np.round(maskArea,4)}', 'red',font=font)
                PIL_image.save(os.path.join(ImagePath,image_name + '_' + f'{idx}.png'))
        dfGran = pd.DataFrame(dfGran)
        dfGran['label'] = 0
        dfGran.to_csv(os.path.join(DataPath,'imageseq_data_area.csv'))
        
    


#%%
import dash
from dash import html
from dash import dcc
from dash.dependencies import Input, Output
import dash_core_components as dcc
import dash_bootstrap_components as dbc
from dash import State
import plotly.express as px
import pandas as pd
import sys
import numpy as np
import matplotlib.pyplot as plt
from PIL import Image
import seaborn as sns
import os
import re
import glob
import random
from sklearn import preprocessing
from tqdm import tqdm
import plotly.express as px
from PIL import Image
import pandas as pd
import plotly.figure_factory as ff
from AIPyS.supportFunctions.AIPS_functions import areaThreshold

# dataPath = r'D:\Gil\images\pex_project\02142024\inc_origImage\imageSequence\data'
# imagePath = r'D:\Gil\images\pex_project\02142024\inc_origImage\imageSequence\images'

def area_viz(imagePath,dataPath):
    df = pd.read_csv(os.path.join(dataPath,'imageseq_data_area.csv'))
    # MLE estimation of area threshold
    th = np.round(areaThreshold(arr_area = df.maskArea.values),3)
    #float_cols = df.select_dtypes(include=['float']).columns
    #df[float_cols] = df[float_cols].apply(lambda x: np.round(x, 3))
    np.random.seed(42423)
    #df = df.sample(frac=1).reset_index(drop=True)
    #df['name'] = df['name'].apply(lambda x: f"{x}.png")
    first_image_name = df.iloc[0]['name']
    first_image_path = f"{imagePath}/{first_image_name}"
    savedContainer = {'name':[], 'intensity':[], 'sd':[], 'maskArea':[]}
    image_names = [re.sub("_.*","",name) for name in set(df.name.tolist())]
    # lookuptableFunction:
    def lookUp(hover,columnName):
        name = df.loc[df[columnName]==hover,columnName]
        return name    

    def openImageToPX(imageName):
        im_pil = Image.open(imageName)
        img = px.imshow(im_pil, binary_string=True, binary_backend="png", width=500, height=500,binary_compression_level=9).update_xaxes(showticklabels=False).update_yaxes(showticklabels = False)
        return img

    # display distplot
    hist_data = [df.loc[:,'maskArea'].values]
    image_name_txt = [df.loc[:,'name'].values]
    group_labels = ['area']
    fig = ff.create_distplot(hist_data, group_labels, bin_size=len(hist_data)/10,rug_text = image_name_txt)

    app = dash.Dash(__name__,prevent_initial_callbacks=True,suppress_callback_exceptions=True)
    application = app.server
    app.layout = html.Div([
        html.Div([
            dbc.Alert("Select images from rug plot:", color="primary"),
            dcc.Dropdown(["maskArea"],placeholder='Plot type', id='select-col',value = "maskArea"),
            dcc.Graph(id="graph-dis-dcc", figure=fig, clear_on_unhover=True),
        ],style={'flex': '1', 'display': 'flex', 'flex-direction': 'column'}),
        html.Div([
            dcc.Loading(html.Div(id='img-container'),type="circle", style={'height': '100%', 'width': '100%'})
            ], style={'flex': '1'}),
        ],style={'display': 'flex', 'flex-direction': 'row'})

    @app.callback(
        Output("graph-dis-dcc", "figure"),
        Input("select-col", "value"))
    def update_plot(col):
        hist_data = [df.loc[:,col].values]
        image_name_txt = [df.loc[:,'name'].values]
        group_labels = ['phno', 'wt']
        fig = ff.create_distplot(hist_data, group_labels, bin_size=len(hist_data)/10,rug_text = image_name_txt)
        fig.add_vline(x=th, line_width=3, line_dash="dash", line_color="red")
        return fig

    @app.callback(
        Output("img-container", "children"),
        Input("graph-dis-dcc", "hoverData"))
    def display_hover(hoverData):
        if hoverData is None:
            return dash.no_update
        if 'text' in [x for x in hoverData["points"][0]]:
            img_name = hoverData["points"][0]["text"]
            if img_name is None:
                return dash.no_update
            img = openImageToPX(os.path.join(imagePath,img_name))
            return [dcc.Graph(id="disp3",figure=img)]
        return dash.no_update
    return app
    


import numpy as np
from skimage.filters import threshold_local
from skimage.util import img_as_ubyte
from scipy.ndimage.morphology import binary_opening
import skimage.morphology as sm
from skimage.segmentation import watershed
from skimage import measure
import os


from AIPyS.DataLoad import DataLoad

class GlobalSeg(DataLoad):
    '''
    Initiate the Global segmentation
    Parameters
    ----------
        Image_name: (str)
        path: (str)
        block_size_cyto: int
            Detect local edges 1-99 odd
        offset_cyto: float
            Detect local edges 0.001-0.9 odd
        global_ther: float
            Percentile
        clean : int
            opening operation, matrix size must be odd
    
    Returns
    -------
        combine: img
            global threshold binary map (eg cytoplasm)
    '''
    # def __init__(self,Image_name,outPath):
    #     super().__init__(Image_name,outPath)
    def __init__(self,Image_name,outPath,block_size_cyto = 11, offset_cyto = 0.1, global_ther = 0.1, clean  = 5,channel = 2,bit8 = 255,ci = 1):
        self.block_size_cyto = block_size_cyto
        self.offset_cyto = offset_cyto
        self.global_ther = global_ther
        self.clean = clean
        self.channel = channel
        self.bit8 = bit8
        self.ci = ci
        super().__init__(Image_name,outPath)
        
    def cytosolSegmentation(self,image):
        # load matrix from seedSegmentation module
        ch2 = image
        ther_cell = threshold_local(ch2, self.block_size_cyto, "gaussian", self.offset_cyto)
        blank = np.zeros_like(ch2)
        cell_mask_1 = ch2 > ther_cell
        cell_mask_2 = binary_opening(cell_mask_1, structure=np.ones((self.clean,self.clean))).astype(np.float64)
        # global threshold
        quntile_num = np.quantile(ther_cell, self.global_ther)
        cell_mask_3 = np.where(ther_cell > quntile_num, 1, 0)
        # in-case of segmentation failed
        if np.sum(cell_mask_3)==0:
            outTarget["cell_mask_1"] = cell_mask_2
            self.out_target = outTarget
            return  self.out_target
        combine = cell_mask_2
        combine[cell_mask_3 > combine] = cell_mask_3[cell_mask_3 > combine]
        return combine

    def disply_compsite(self, image, mask):
        '''
        bit8 color assign 255 is very bright
        '''
        input_gs_image = image
        input_gs_image = input_gs_image*self.ci
        input_gs_image = (input_gs_image / input_gs_image.max()) * 255
        ch2_u8 = np.uint8(input_gs_image)
        rgb_input_img = np.zeros((np.shape(ch2_u8)[0], np.shape(ch2_u8)[1], 3), dtype=np.uint8)
        rgb_input_img[:, :, 0] = ch2_u8
        rgb_input_img[:, :, 1] = ch2_u8
        rgb_input_img[:, :, 2] = ch2_u8
        rgb_input_img[mask > 0, self.channel] = self.bit8
        return rgb_input_img

    def rgbMasking(self,image):
        mask = self.cytosolSegmentation(image = image)
        frame = self.disply_compsite(image = image, mask = mask)
        return frame,mask

    def openingOperation_(self,kernel,image):
        '''
        Parameters
        ---------- 
            kernel: int, 
                size of filter kernel
        return
        ------
            opening operation image
        '''
        selem = morphology.disk(kernel, dtype=bool)
        eros_pix = morphology.erosion(image, footprint=selem)
        imageOpen = morphology.dilation(eros_pix, footprint=selem)
        return imageOpen
    



import os
import shutil
import h5py
from pathlib import Path
import numpy as np
import argparse
import re
import pdb
from pathlib import Path

def main(path):
    newpath = path.encode('utf-8')
    print(newpath)

if __name__ == "__main__":
    parser = argparse.ArgumentParser(description='Update or add parameters and display the content of the parameters.h5 file. This script allows users to manage and view parameters stored in a .h5 file. Users can update parameter values or simply display the current configuration.', formatter_class=argparse.RawTextHelpFormatter)
    # Adding detailed help messages for each argument
    parser.add_argument('--data_dir', type=str, help="Directory where data is stored.\nThis path is used to locate your data files for processing.")
    args = parser.parse_args()
    main(args.data_dir)
    


import numpy as np
from skimage.util import img_as_ubyte
import skimage.morphology as sm
from skimage.segmentation import watershed
import os
import cv2
from AIPyS.supportFunctions.AIPS_file_display import Compsite_display
from AIPyS.segmentation.cellpose.AIPS_cellpose import AIPS_cellpose
from AIPyS.supportFunctions.GranularityFunc import resize

class CellPoseSeg(AIPS_cellpose):
    def __init__(self,videoName,*args, **kwargs):
        self.videoName = videoName
        super().__init__(*args, **kwargs)
        self.imageSeqGen_cp()
    
    def check_multiImage(self):
        if len(self.img) < 2:
            assert "list of images are required"
    
    def imageSeqGen_cp(self):
        self.check_multiImage() # except list of images
        images = self.img
        mask, _ = self.cellpose_segmentation(image_input = images[0])
        frame_obj = Compsite_display(input_image = images[0], mask_roi = mask, channel =  2)
        frame = resize(frame_obj.rgb_out, 500,500)
        width, height, layers = frame.shape
        video = cv2.VideoWriter(os.path.join(self.outPath, self.videoName), cv2.VideoWriter_fourcc(*'MJPG'), 1.0,(500, 500), isColor=True)
        print('analysis might take a few minutes')
        for i, image in enumerate(images):
            print('-' * i, end='\r')
            mask, _ = self.cellpose_segmentation(image_input=image)
            frame_obj = Compsite_display(input_image = image, mask_roi = mask, channel =  2)
            frame = resize(frame_obj.rgb_out, 500,500)
            video.write(frame)
        video.release()
        print(f'{self.videoName} is saved')









'''
    Deployment of convolutional neural network
'''
import time, os, sys
from tensorflow.keras.preprocessing.image import ImageDataGenerator, load_img, img_to_array, array_to_img
from tensorflow.keras.models import load_model
import matplotlib.pyplot as plt
import numpy as np
import tifffile as tfi
from AIPyS.standAlone import AIPS_cellpose as AC

def CNNDeploy(path_model,model,file_name,path,pathOut,areaFilter,thr=0.5):
    """
    Parameters
    ----------
    path_model: str
    model: h5
        tensor flow model file
    file_name: str
    path: str
    pathOut: str
    areaFilter: int
    thr: float
    """
    model_cnn = load_model(os.path.join(path_model, model))
    AIPS_pose_object = AC.AIPS_cellpose(Image_name=file_name, path=path, model_type="cyto", channels=[0, 0])
    img = AIPS_pose_object.cellpose_image_load()
    # create mask for the entire image
    mask, table = AIPS_pose_object.cellpose_segmantation(image_input=img)
    table["predict"] = 'Na'
    for i in range(len(table)):
        stack, stack_v = AIPS_pose_object.stackObjects_cellpose_ebimage_parametrs_method(image_input=img,
                                                                                         extract_pixel=50,
                                                                                         resize_pixel=150,
                                                                                         img_label=table.index.values[
                                                                                             i])
        if stack is None:
            continue
        else:
            plt.imsave("temp.png", stack)
            test_imgs = img_to_array(load_img("temp.png", target_size=(150, 150)))
            test_imgs = np.array(test_imgs)
            test_imgs_scaled = test_imgs
            test_imgs_scaled /= 255
            pred = model_cnn.predict(test_imgs_scaled.reshape(1, 150, 150, 3), verbose=0).tolist()[0][0]
            table.loc[i, "predict"] = pred
    # remove nas
    table_na_rmv = table.loc[table['predict'] != 'Na', :]
    table_na_rmv = table_na_rmv.loc[table['area'] > areaFilter, :]
    ##### binary image contains the phnotype of intrse #####
    image_blank = np.zeros((np.shape(img)[0], np.shape(img)[1]))
    binary, table_sel = AIPS_pose_object.call_bin(table_sel_cor=table_na_rmv, threshold=thr, img_blank=image_blank)
    tfi.imsave(os.path.join(pathOut, 'binary.tif'), binary)


from AIPyS.CLI.promptParameters import check_and_prompt_parameters

def parametersInspec(option, user_parameters_path):
    parameters,non_counts,missingParam = check_and_prompt_parameters(option, user_parameters_path)
    if non_counts > 0:
        print(f" {subkey} is required. use set_parameters --{subkey} to update parameters" for subkey in missingParam)
        return
    else:
        return parameters
